#include <string.h>
#include <stdlib.h>
#include "disasm.h"

char const * CPadaukDisasm:: ak(void) { sprintf(sa, "0x%02X", opcode & 0x00FF); return sa; }
char const * CPadaukDisasm::aak(void) { sprintf(sa, "a, 0x%02X", opcode & 0x00FF); return sa; }
char const * CPadaukDisasm::aam(void) { sprintf(sa, "a, %s", mem_symbols[get_m()]); return sa; }
char const * CPadaukDisasm::ama(void) { sprintf(sa, "%s, a", mem_symbols[get_m()]); return sa; }
char const * CPadaukDisasm:: am(void) { sprintf(sa, "%s", mem_symbols[get_m()]); return sa; }
char const * CPadaukDisasm:: aw(void) { sprintf(sa, "%s", mem_symbols[get_w()]); return sa; }
char const * CPadaukDisasm::awa(void) { sprintf(sa, "[%s],a", mem_symbols[get_w()]); return sa; }
char const * CPadaukDisasm::aaw(void) { sprintf(sa, "a, [%s]", mem_symbols[get_w()]); return sa; }
char const * CPadaukDisasm::aap(void) { sprintf(sa, "a, %s", io_symbols[get_p()]); return sa; }
char const * CPadaukDisasm::apa(void) { sprintf(sa, "%s, a", io_symbols[get_p()]); return sa; }
char const * CPadaukDisasm::apb(void) { sprintf(sa, "%s.%s", io_symbols[get_p()], io_symbols.bit(get_p(), get_b())); return sa; }
char const * CPadaukDisasm::amb(void) {
	unsigned b, m;
	get_b_m(b, m);
	sprintf(sa, "%s.%s", mem_symbols[m], mem_symbols.bit(m, b));
	return sa;
}
char const * CPadaukDisasm:: aj(void) { sprintf(sa, "0x%04X", get_j()); return sa; }
char const * CPadaukDisasm:: af(void) { sprintf(sa, "pc%i", get_f()); return sa; }
char const * CPadaukDisasm:: ad(void) { sprintf(sa, "%i", get_d()); return sa; }


struct CPadaukDisasm::TAR CPadaukDisasm::args[] = {
	//								      --- 16 bit core ---	   --- 14 bit core ---	     --- 13 bit core ---
	//								     64 byte IO  ( 6 bits)	  64 byte IO  ( 6 bits)	    32 byte IO  ( 5 bits)
	//								    512 byte RAM ( 9 bits)	 128 byte RAM ( 7 bits)	   128 byte RAM ( 7 bits)
	//								    8kw program  (13 bits)	 2kw program  (11 bits)	   1kw program  (10 bits)
	{ {   0,   0,   0, 0 },	 "",		&CPadaukDisasm::a_,	{ { "             ", 0x0000 }, { "           ", 0x0000 }, { "          ", 0x0000 } } },
	{ { 'a',   0,   0, 0 },  "a",		&CPadaukDisasm::aa,	{ { "             ", 0x0000 }, { "           ", 0x0000 }, { "          ", 0x0000 } } },
	{ { 'k',   0,   0, 0 },  "I",		&CPadaukDisasm::ak,	{ { "     kkkkkkkk", 0x00FF }, { "   kkkkkkkk", 0x00FF }, { "  kkkkkkkk", 0x00FF } } },
	{ { 'a', ',', 'k', 0 },  "a,I",		&CPadaukDisasm::aak,	{ { "     kkkkkkkk", 0x00FF }, { "   kkkkkkkk", 0x00FF }, { "  kkkkkkkk", 0x00FF } } },
	{ { 'a', ',', 'm', 0 },  "a,M",		&CPadaukDisasm::aam,	{ { "    mmmmmmmmm", 0x01FF }, { "    mmmmmmm", 0x007F }, { "    mmmmmm", 0x003F } } },
	{ { 'm', ',', 'a', 0 },  "M,a",		&CPadaukDisasm::ama,	{ { "    mmmmmmmmm", 0x01FF }, { "    mmmmmmm", 0x007F }, { "    mmmmmm", 0x003F } } },
	{ { 'm',   0,   0, 0 },  "M",		&CPadaukDisasm::am,	{ { "    mmmmmmmmm", 0x01FF }, { "    mmmmmmm", 0x007F }, { "    mmmmmm", 0x003F } } },
	{ { 'w',   0,   0, 0 },  "word",	&CPadaukDisasm::aw,	{ { "    wwwwwwww ", 0x01FE }, { "    wwwwww ", 0x007E }, { "     wwww ", 0x001E } } },
	{ { 'w', ',', 'a', 0 },  "index,a",	&CPadaukDisasm::awa,	{ { "    wwwwwwww ", 0x01FE }, { "    wwwwww ", 0x007E }, { "     wwww ", 0x001E } } },
	{ { 'a', ',', 'w', 0 },  "a,index",	&CPadaukDisasm::aaw,	{ { "    wwwwwwww ", 0x01FE }, { "    wwwwww ", 0x007E }, { "     wwww ", 0x001E } } },
	{ { 'a', ',', 'p', 0 },  "a,IO",	&CPadaukDisasm::aap,	{ { "       pppppp", 0x003F }, { "     pppppp", 0x003F }, { "     ppppp", 0x001F } } },
	{ { 'p', ',', 'a', 0 },  "IO,a",	&CPadaukDisasm::apa,	{ { "       pppppp", 0x003F }, { "     pppppp", 0x003F }, { "     ppppp", 0x001F } } },
	{ { 'p', '.', 'b', 0 },  "IO.n",	&CPadaukDisasm::apb,	{ { "    bbbpppppp", 0x01FF }, { "  bbbpppppp", 0x01FF }, { "  bbbppppp", 0x00FF } } },
	{ { 'm', '.', 'b', 0 },  "M.n",		&CPadaukDisasm::amb,	{ { " bbbmmmmmmmmm", 0x0FFF }, { "  bbbmmmmmm", 0x01FF }, { "  bbb mmmm", 0x00EF } } },
	{ { 'j',   0,   0, 0 },  "label",	&CPadaukDisasm::aj,	{ { "aaaaaaaaaaaaa", 0x1FFF }, { "aaaaaaaaaaa", 0x07FF }, { "aaaaaaaaaa", 0x03FF } } },
	{ { 'f',   0,   0, 0 },  "pcN",		&CPadaukDisasm::af,	{ { "         tttt", 0x000F }, { "           ",      0 }, { "          ",      0 } } },
	{ { 'd',   0,   0, 0 },  "n",		&CPadaukDisasm::ad,	{ { "        kkkkk", 0x001F }, { "           ",      0 }, { "          ",      0 } } },
	{ {   0,   0,   0, 0 },  0,				NULL,	{ {               0,      0 }, {             0,      0 }, {            0,      0 } } }
};

CPadaukDisasm::TOP  CPadaukDisasm::ops[] = {
	{ "???",	"",	"    ", { 0x0000, 0x0000, 0x0000 },	"Invalid opcode" },
	{ "nop",	"",	"    ", { 0x0000, 0x0000, 0x0000 },	"No operation" },
	//	{ "trap",	"",		"    ",	{ 0x0000, 0x0000, 0x0000 },	"Used only by ICE" },
	{ "ldsptl",	"",	"????", { 0x0000, 0x0006, 0x0000 },	"A <- [SP]" },
	{ "ldspth",	"",	"????", { 0x0000, 0x0007, 0x0000 },	"A <- [SP + 1]" },
	{ "addc",	"a",	"ZCAO", { 0x0010, 0x0060, 0x0010 },	"A <- A + CF" },
	{ "subc",	"a",	"ZCAO", { 0x0011, 0x0061, 0x0011 },	"A <- A - CF" },
	{ "izsn",	"a",	"ZCAO", { 0x0012, 0x0062, 0x0012 },	"Increment A and skip next instruction if A is zero" },
	{ "dzsn",	"a",	"ZCAO", { 0x0013, 0x0063, 0x0013 },	"Decrement A and skip next instruction if A is zero" },
	{ "pcadd",	"a",	"    ", { 0x0017, 0x0067, 0x0017 },	"Add A to PC" },
	{ "not",	"a",	"Z   ", { 0x0018, 0x0068, 0x0018 },	"A <- ~A" },
	{ "neg",	"a",	"Z   ", { 0x0019, 0x0069, 0x0019 },	"A <- NEG(A)" },
	{ "sr",		"a",	" C  ", { 0x001A, 0x006A, 0x001A },	"A <- A >> 1" },
	{ "sl",		"a",	" C  ", { 0x001B, 0x006B, 0x001B },	"A <- A << 1" },
	{ "src",	"a",	" C  ", { 0x001C, 0x006C, 0x001C },	"A <- CF : A >> 1" },
	{ "slc",	"a",	" C  ", { 0x001D, 0x006D, 0x001D },	"A <- A : CF << 1" },
	{ "swap",	"a",	"    ", { 0x001E, 0x006E, 0x001E },	"Swap the high nibble and low nibble of A" },
	{ "delay",	"a",	"    ", { 0x001F, 0x0000, 0x0000 },	"" },
	{ "wdreset","",		"    ", { 0x0030, 0x0070, 0x0030 },	"Reset Watchdog timer" },
	{ "pushaf",	"",	"    ", { 0x0032, 0x0072, 0x0032 },	"Push A and flags to stack" },
	{ "popaf",	"",	"ZCAO", { 0x0033, 0x0073, 0x0033 },	"Pop A and flags from stack" },
	{ "reset",	"",	"    ", { 0x0035, 0x0075, 0x0035 },	"Reset the whole chip" },
	{ "stopsys","",		"    ", { 0x0036, 0x0076, 0x0036 },	"System halt (OSC disabled)" },
	{ "stopexe","",		"    ", { 0x0037, 0x0077, 0x0037 },	"CPU halt (OSC active to output clock, SYSCLK disabled to save power)" },
	{ "engint",	"",	"    ", { 0x0038, 0x0078, 0x0038 },	"Global interrupt enbale" },
	{ "disgint","",		"    ", { 0x0039, 0x0079, 0x0039 },	"Global interrupt disable" },
	{ "ret",	"",	"    ", { 0x003A, 0x007A, 0x003A },	"Return from subroutine" },
	{ "reti",	"",	"    ", { 0x003B, 0x007B, 0x003B },	"Return from interrupt" },
	{ "mul",	"",	"    ", { 0x003C, 0x007C, 0x0000 },	"Multiply (if available)" },
	{ "pmode",	"d",	"    ", { 0x0040, 0x0000, 0x0000 },	"" },
	{ "popw",	"f",	"    ", { 0x0060, 0x0000, 0x0000 },	"" },
	{ "pushw",	"f",	"    ", { 0x0070, 0x0000, 0x0000 },	"" },
	{ "mov",	"p,a",	"    ", { 0x0080, 0x0180, 0x0080 },	"IO <- A" },
	{ "mov",	"a,p",	"Z   ", { 0x00C0, 0x01C0, 0x00A0 },	"A <- IO" },
	{ "cneqsn",	"a,k",	"ZCAO", { 0x0100, 0x2B00, 0x1300 },	"Skip next instruction if A not equals k" },
	{ "stt16",	"w",	"    ", { 0x0200, 0x0300, 0x00C0 },	"M <- Timer16 (last bit of M set to 0, M must be word aligned)" },
	{ "ldt16",	"w",	"    ", { 0x0201, 0x0301, 0x00C1 },	"Timer16 <- M last bit of M set to 1, M must be word aligned)" },
	{ "popw",	"w",	"    ", { 0x0400, 0x0000, 0x0000 },	"" },
	{ "pushw",	"w",	"    ", { 0x0401, 0x0000, 0x0000 },	"" },
	{ "igoto",	"w",	"    ", { 0x0600, 0x0000, 0x0000 },	"" },
	{ "icall",	"w",	"    ", { 0x0601, 0x0000, 0x0000 },	"" },
	{ "idxm",	"w,a",	"    ", { 0x0800, 0x0380, 0x00E0 },	"[M] <- A (last bit of M set to 0, M must be word aligned, 2 cycles)" },
	{ "idxm",	"a,w",	"    ", { 0x0801, 0x0381, 0x00E1 },	"A <-[M] (last bit of M set to 1, M must be word aligned, 2 cycles)" },
	{ "ldtabl",	"w",	"    ", { 0x0A00, 0x0000, 0x0000 },	"" },
	{ "ldtabh",	"w",	"    ", { 0x0A01, 0x0000, 0x0000 },	"" },
	{ "delay",	"k",	"    ", { 0x0E00, 0x0000, 0x0000 },	"" },
	{ "ret",	"k",	"    ", { 0x0F00, 0x0200, 0x0100 },	"A <- k and return from subroutine" },
	{ "xor",	"p,a",	"    ", { 0x1000, 0x00C0, 0x0060 },	"IO <- IO ^ A" },
	{ "xor",	"a,p",	"Z   ", { 0x1040, 0x0000, 0x0000 },	"" },
	{ "cneqsn",	"m,a",	"ZCAO", { 0x1400, 0x0000, 0x0000 },	"Skip next instruction if M is not equal to A" },
	{ "cneqsn",	"a,m",	"ZCAO", { 0x1600, 0x1780, 0x0BC0 },	"Skip next instruction if M is not equal to A" },
	{ "add",	"a,k",	"ZCAO", { 0x1800, 0x2800, 0x1000 },	"A <- A + k" },
	{ "sub",	"a,k",	"ZCAO", { 0x1900, 0x2900, 0x1100 },	"A <- A - k" },
	{ "ceqsn",	"a,k",	"ZCAO", { 0x1A00, 0x2A00, 0x1200 },	"Skip next instruction if A equals k" },
	{ "comp",	"a,k",	"ZCAO", { 0x1B00, 0x0000, 0x0000 },	"" },
	{ "and",	"a,k",	"Z   ", { 0x1C00, 0x2C00, 0x1400 },	"A <- A & k" },
	{ "or",		"a,k",	"Z   ", { 0x1D00, 0x2D00, 0x1500 },	"A <- A | k" },
	{ "xor",	"a,k",	"Z   ", { 0x1E00, 0x2E00, 0x1600 },	"A <- A ^ k" },
	{ "mov",	"a,k",	"    ", { 0x1F00, 0x2F00, 0x1700 },	"A <- k" },
	{ "t0sn",	"p.b",	"    ", { 0x2000, 0x1800, 0x0C00 },	"Test bit n of IO and skip next instruction if clear" },
	{ "t1sn",	"p.b",	"    ", { 0x2200, 0x1A00, 0x0D00 },	"Test bit n of IO and skip next instruction if set" },
	{ "set0",	"p.b",	"    ", { 0x2400, 0x1C00, 0x0E00 },	"Clear bit n of IO" },
	{ "set1",	"p.b",	"    ", { 0x2600, 0x1E00, 0x0F00 },	"Set bit n of IO" },
	{ "tog",	"p.b",	"    ", { 0x2800, 0x0000, 0x0000 },	"" },
	{ "wait0",	"p.b",	"    ", { 0x2A00, 0x0000, 0x0000 },	"" },
	{ "wait1",	"p.b",	"    ", { 0x2C00, 0x0000, 0x0000 },	"" },
	{ "swapc",	"p.b",	" C  ", { 0x2E00, 0x0400, 0x0000 },	"Swap bit IO.n with CF" },
	{ "nmov",	"m,a",	"    ", { 0x3000, 0x0000, 0x0000 },	"" },
	{ "nmov",	"a,m",	"    ", { 0x3200, 0x0000, 0x0000 },	"" },
	{ "nadd",	"m,a",	"ZCAO", { 0x3400, 0x0780, 0x0000 },	"M <- NEG(M) + A" },
	{ "nadd",	"a,m",	"ZCAO", { 0x3600, 0x0700, 0x0000 },	"A <- M + NEG(A)" },
	{ "ceqsn",	"m,a",	"ZCAO", { 0x3800, 0x0000, 0x0000 },	"Skip next instruction if M is equal to A" },
	{ "ceqsn",	"a,m",	"ZCAO", { 0x3A00, 0x1700, 0x0B80 },	"Skip next instruction if M is equal to A" },
	{ "comp",	"m,a",	"ZCAO", { 0x3C00, 0x0680, 0x0000 },	"Compare M with A (flags changed according to (M - A))" },
	{ "comp",	"a,m",	"ZCAO", { 0x3E00, 0x0600, 0x0000 },	"Compare A with M (flags changed according to (A - M))" },
	{ "add",	"m,a",	"ZCAO", { 0x4000, 0x0800, 0x0400 },	"M <- M + A" },
	{ "add",	"a,m",	"ZCAO", { 0x4200, 0x0C00, 0x0600 },	"A <- A + M" },
	{ "sub",	"m,a",	"ZCAO", { 0x4400, 0x0880, 0x0440 },	"M <- M - A" },
	{ "sub",	"a,m",	"ZCAO", { 0x4600, 0x0C80, 0x0640 },	"A <- A - M" },
	{ "addc",	"m,a",	"ZCAO", { 0x4800, 0x0900, 0x0480 },	"M <- M + A + CF" },
	{ "addc",	"a,m",	"ZCAO", { 0x4A00, 0x0D00, 0x0680 },	"A <- A + M + CF" },
	{ "subc",	"m,a",	"ZCAO", { 0x4C00, 0x0980, 0x04C0 },	"M <- M - A - CF" },
	{ "subc",	"a,m",	"ZCAO", { 0x4E00, 0x0D80, 0x06C0 },	"A <- A - M - CF" },
	{ "and",	"m,a",	"Z   ", { 0x5000, 0x0A00, 0x0500 },	"M <- M & A" },
	{ "and",	"a,m",	"Z   ", { 0x5200, 0x0E00, 0x0700 },	"A <- A & M" },
	{ "or",		"m,a",	"Z   ", { 0x5400, 0x0A80, 0x0540 },	"M <- M | A" },
	{ "or",		"a,m",	"Z   ", { 0x5600, 0x0E80, 0x0740 },	"A <- A | M" },
	{ "xor",	"m,a",	"Z   ", { 0x5800, 0x0B00, 0x0580 },	"M <- M ^ A" },
	{ "xor",	"a,m",	"Z   ", { 0x5A00, 0x0F00, 0x0780 },	"A <- A ^ M" },
	{ "mov",	"m,a",	"    ", { 0x5C00, 0x0B80, 0x05C0 },	"M <- A" },
	{ "mov",	"a,m",	"Z   ", { 0x5E00, 0x0F80, 0x07C0 },	"A <- M" },
	{ "addc",	"m",	"ZCAO", { 0x6000, 0x1000, 0x0800 },	"M <- M + CF" },
	{ "subc",	"m",	"ZCAO", { 0x6200, 0x1080, 0x0840 },	"M <- M - CF" },
	{ "izsn",	"m",	"ZCAO", { 0x6400, 0x1100, 0x0880 },	"M <- M + 1, skip next instruction if M is 0" },
	{ "dzsn",	"m",	"ZCAO", { 0x6600, 0x1180, 0x08C0 },	"M <- M - 1, skip next instruction if M is 0" },
	{ "inc",	"m",	"ZCAO", { 0x6800, 0x1200, 0x0900 },	"M <- M + 1" },
	{ "dec",	"m",	"ZCAO", { 0x6A00, 0x1280, 0x0940 },	"M <- M - 1" },
	{ "clear",	"m",	"    ", { 0x6C00, 0x1300, 0x0980 },	"M <- 0" },
	{ "xch",	"m",	"    ", { 0x6E00, 0x1380, 0x09C0 },	"Exchange A with M" },
	{ "not",	"m",	"Z   ", { 0x7000, 0x1400, 0x0A00 },	"M <- ~M" },
	{ "neg",	"m",	"Z   ", { 0x7200, 0x1480, 0x0A40 },	"M <- NEG(M)" },
	{ "sr",		"m",	" C  ", { 0x7400, 0x1500, 0x0A80 },	"M <- M >> 1" },
	{ "sl",		"m",	" C  ", { 0x7600, 0x1580, 0x0AC0 },	"M <- M << 1" },
	{ "src",	"m",	" C  ", { 0x7800, 0x1600, 0x0B00 },	"M <- CF : M >> 1" },
	{ "slc",	"m",	" C  ", { 0x7A00, 0x1680, 0x0B40 },	"M <- M : CF << 1" },
	{ "swap",	"m",	"    ", { 0x7C00, 0x0000, 0x0000 },	"" },
	{ "delay",	"m",	"    ", { 0x7E00, 0x0000, 0x0000 },	"" },
	{ "t0sn",	"m.b",	"    ", { 0x8000, 0x2000, 0x0200 },	"Test bit n of memory M and skip next instruction if clear" },
	{ "t1sn",	"m.b",	"    ", { 0x9000, 0x2200, 0x0210 },	"Test bit n of memory M and skip next instruction if set" },
	{ "set0",	"m.b",	"    ", { 0xA000, 0x2400, 0x0300 },	"Clear bit n of memory M" },
	{ "set1",	"m.b",	"    ", { 0xB000, 0x2600, 0x0310 },	"Set bit n of memory M" },
	{ "goto",	"j",	"    ", { 0xC000, 0x3000, 0x1800 },	"Jump" },
	{ "call",	"j",	"    ", { 0xE000, 0x3800, 0x1C00 },	"Call subroutine" },
	{ 0,		0,	0,	{ 0xFFFF, 0xFFFF, 0xFFFF }, 0 }
};


CPadaukDisasm::TAR const * CPadaukDisasm::find_arg(char const * const s)
{
	auto const *a = args;
	for (; a->pargs; ++a) if (!strcmp(a->args, s)) break;
	return  a;
}

 void CPadaukDisasm::patch_opcode_table(void)
{
	if (ops[0].args.p == args) return;
	for (auto *o = ops; o->mnemonic; ++o) o->args.p = find_arg(o->args.s);
}

 
void CPadaukDisasm::init_disasm(unsigned const e)
{
	patch_opcode_table();
	auto encoding = e <= 2 ? e : 0;
	memset(oi, 0, sizeof(oi));
	for (auto const *o = ops + 1; o->mnemonic; ++o) {
		auto const op = o->opcode[e];
		if (op || o == ops + 1) {
			auto const m = o->args.p->e[e].mask;
			for (unsigned n = 0; n <= m; ++n)
				if ((n & m) == n) oi[op + n] = o - ops;
		}
	}
}

void CPadaukDisasm::disasm(uint16_t const w)
{
	opcode = w;
	auto const * const o = ops + oi[opcode];
	printf("%-10s%s\n", o->mnemonic, (this->*(o->args.p->make_arg))());
}

void CPadaukDisasm::disasm(CPDK &pdk)
{
	init_disasm();

	auto *w = pdk.pdata;
	for (unsigned n = 0; n < pdk.length; ++n, ++w) {
	//for (unsigned n = 0; n < 0x00E0; ++n, ++w) {
		printf("%04X  %04X  ", n, *w);
		disasm(*w);
	}
}

 char const * CPadaukDisasm::instruction_encoding(unsigned const size, unsigned const opcode, char const * const bits)
{
	if (size < 1 || size > 16) return "size_out_of_range";

	unsigned o = strlen(bits);
	if (o > size) return "too_many_bits";
	o = size - o;

	unsigned m = (1 << (size - 1));
	static char e[20];
	char *c = e;
	do {
		*c++ = (opcode & m) ? '1' : '0';
		if (m & 0x1110) *c++ = '_';
	} while (m >>= 1);
	*c = 0;

	char b[17];
	memset(b, ' ', o);
	strcpy(b, bits);
	char const *r = b;
	c = e;
	do {
		if (*c == '_') ++c;
		if (*r != ' ') *c = *r;
		++r;
	} while (*++c);

	return e;
}

void CPadaukDisasm::print_instruction_set(unsigned e)
{
	patch_opcode_table();

	unsigned const eb[3] = { 16, 14, 13 };
	unsigned const bits = eb[e];

	auto const *o = ops + 1;

	while (o->mnemonic) {
		auto const op = o->opcode[e];

		printf("0x%04X  %s  %-9s%s\n",
			op,
			instruction_encoding(bits, op, o->args.p->e[e].bits),
			o->mnemonic,
			o->args.p->pargs);

		while ((++o)->mnemonic && !o->opcode[e]);
	}
}

void CPadaukDisasm::compare_opcodes(void)
{
	patch_opcode_table();

	for (auto const *o = ops + 1; o->mnemonic; ++o) {
		printf("%-10s", o->mnemonic);
		printf("%-10s", o->args.p->pargs);
		const bool nop = o == ops + 1;
		printf("%-22s", (o->opcode[0] || nop) ? instruction_encoding(16, o->opcode[0], o->args.p->e[0].bits) : "");
		printf("%-20s", (o->opcode[1] || nop) ? instruction_encoding(14, o->opcode[1], o->args.p->e[1].bits) : "");
		printf("%-16s", (o->opcode[2] || nop) ? instruction_encoding(13, o->opcode[2], o->args.p->e[2].bits) : "");
		printf("\n");
	}
}

CPadaukDisasm * CPadaukDisasm::New(unsigned bits)
{
	switch (bits) {
		case 16: return new CPadaukDisasm16;
		case 14: return new CPadaukDisasm14;
		case 13: return new CPadaukDisasm13;
		default: return reinterpret_cast<CPadaukDisasm *>(0);
	}
}
