#include "symbols.h"
#include <fstream>
#include <string>
#include <string.h>
#include <cstdarg>

int CSymbols::ReadFile(char const * const file)
{
	std::ifstream is(file);
	if (!is.is_open()) {
		printf("Unable to open symbol file %s. Using default symbols.\n", file);
		return -1;
	}
	std::string line;
	while (std::getline(is, line)) {
		if (!line.length()) continue;
		if (line[0] == ';') continue;
		auto const t = new char[line.length() + 10];
		auto *p = t; *p = 0;
		unsigned n = 9;
		for(auto &c : line) {
			if (iswspace(c)) {
				if (*p) {
					*++p = 0;
					if (!--n) break;
				}
			} else {
				*++p = c;
			}
		}
		for (; n; --n) *++p = 0;
		unsigned const l = p - t;
		auto s = new char[l];
		memcpy(s, t + 1, l);
		delete t;
		symbols.push_back(s);
	}
	return 0;
}

char const * CSymbols::operator [] (unsigned const n)
{
	if( n < symbols.size()) return symbols[n];
	static char s[32];
	sprintf(s, "%s%i", default_symbol.c_str(), n);
	return s;
}

char const * CSymbols::bit(char const * s, unsigned const b)
{
	static char const *default_bits[] = { "0", "1", "2", "3", "4", "5", "6", "7" };
	unsigned n = 8 - b;
	do {
		while (*s) ++s;
		++s;
	} while (--n);
	return *s ? s : default_bits[b];
}

char const * CSymbols::bit(unsigned const n, unsigned const b)
{
	static char const default_bits[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	return bit(n < symbols.size() ? symbols[n] : default_bits, b);
}
